package org.processmining.localprocessmodelconformance.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;
import org.processmining.models.graphbased.directed.transitionsystem.TransitionSystem;
import org.processmining.models.semantics.petrinet.CTMarking;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.plugins.petrinet.behavioralanalysis.CGGenerator;
import org.processmining.pnanalysis.metrics.PetriNetMetric;
import org.processmining.pnanalysis.semantics.CyclomaticSemantics;

/*
 * K.B. Lassen and W.M.P. van der Aalst. 
 * Complexity Metrics for Workflow Nets. 
 * Information and Software Technology, 51(3):610-626, 2009.
 */
public class PetriNetCyclomaticMetric implements PetriNetMetric {

	public final static String NAME = "Extended Cyclomatic metric";
	private StringBuffer buffer;
	
	public double compute(PluginContext context, Petrinet net, Marking marking) {
		Map<State, Integer> componentMap = new HashMap<State, Integer>();
		int component = 0;
		int nofStates = Integer.MAX_VALUE;
		buffer = new StringBuffer();
		boolean isBounded = true;
		PetrinetSemantics cyclomaticSemantics = new CyclomaticSemantics(nofStates);
		cyclomaticSemantics.initialize(net.getTransitions(), marking);

		TransitionSystem cg = CGGenerator.getCoverabilityGraph(net, marking, cyclomaticSemantics);

		if (cg.getStates().size() > nofStates) {
			buffer.append("State space too big\n");
			return -1.0;
		}
		for (Object state : cg.getStates()) {
			if (!((CTMarking) state).getOmegaPlaces().isEmpty()) {
				isBounded = false;
			}
		}
		if (!isBounded) {
			buffer.append("Reachability graph is unbounded\n");
			return -1.0;
		}
		for (State state : cg.getNodes()) {
			if (!componentMap.keySet().contains(state)) {
				Set<State> forwardStates = forward(cg, state);
				Set<State> backwardStates = backward(cg, state);
				forwardStates.retainAll(backwardStates);
				component++;
				for (State componentNode : forwardStates) {
					componentMap.put(componentNode, component);
				}
			}
		}
		buffer.append("|E| = " + cg.getEdges().size() + "\n");
		buffer.append("|V| = " + cg.getNodes().size() + "\n");
		buffer.append("p = " + component);
		return cg.getEdges().size() - cg.getNodes().size() + component;
	}
	
	public String getName() {
		return NAME;
	}

	private Set<State> forward(TransitionSystem graph, State startState) {
		Set<State> states = new HashSet<State>();
		List<State> toDo = new ArrayList<State>();
		toDo.add(startState);
		while (!toDo.isEmpty()) {
			State state = toDo.remove(0);
			if (!states.contains(state)) {
				states.add(state);
				for (Transition edge : graph.getOutEdges(state)) {
					toDo.add(edge.getTarget());
				}
			}
		}
		return states;
	}

	private Set<State> backward(TransitionSystem net, State startState) {
		Set<State> states = new HashSet<State>();
		List<State> toDo = new ArrayList<State>();
		toDo.add(startState);
		while (!toDo.isEmpty()) {
			State state = toDo.remove(0);
			if (!states.contains(state)) {
				states.add(state);
				for (Transition edge : net.getInEdges(state)) {
					toDo.add(edge.getSource());
				}
			}
		}
		return states;
	}

	public StringBuffer getLog() {
		return buffer;
	}
}